package orderprocessing;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/*
	History
	date		description
	-----		-----------
	11/12/99	Modifed from the original, create separate address class

*/

/** This class maintains information about a single customer
	<p>
    Note that any changes to the public part of this class or any change in
    the fields will require versioning. This is because these objects are
    stored in the object database for this application.
    <p>
    This class can serve as a model for other classes that need versioning. Be
    sure to add versioning infrastructure to them before you include them in
    the database for the first time.
    <p>
    Note that Addresses are already implicitly included in the database since they
    are written whenever a Customer is written.
    <p>
    Additionally, this class maintains a collection of all customers.
*/
public class Customer implements Serializable
{	/** create a customer object with default values for ship address and terms
	*/
    public Customer(String name, Address billingAddress)
	{	this(name, billingAddress, billingAddress, BillingTerms.get("NET 30"));
	}

	/** Create a customer object with default terms.
	*/
    public Customer(String name, Address billingAddress, Address shipAddress)
	{	// To Do: error checking for id and name
	    this.name      			= name;
	    this.billingAddress		= billingAddress;
	    this.shipAddress 		= shipAddress;
	    key = IDGenerator.nextID();
	    allCustomers.put(key, this);
	}

	/** create a customer object with all options specified.
	* @param name the customer name.
	* @param billingAddress the customer billing address
	* @param shipAddress the ship address if not specified on an order
	* @param billingTerms the billing terms if not specified on an order.
	*/
    public Customer
    (	String name,
    	Address billingAddress,
    	Address shipAddress,
    	BillingTerms billingTerms
    )
	{	this.name       		= name;
	    this.billingAddress		= billingAddress;
	    this.shipAddress 		= shipAddress;
	    this.terms 				= billingTerms;
		key = IDGenerator.nextID();

        allCustomers.put(key, this);
	}


	// Accessors
	public String  key()        { return key;      }
	public String  name()       { return name;    }
	public Address billingAddress()    { return billingAddress; }
	public Address shippingAddress()    { return shipAddress; }
	public BillingTerms billingTerms() { return terms;}

    // Mutators
	public void billingTerms(BillingTerms terms){this.terms = terms;}

	public void name(String name)
	{	this.name = name;
		billingAddress.name(name); //avoid redundancy.
	}

	// other methods
	public void display()
	{	System.out.println(this);
	}

	public boolean equals(Object other)
	{	if(!(other instanceof Customer))return false;
		return key.equals(((Customer)other).key);
	}

	public int hashcode()
	{	return key.hashCode()*997;
	}

	public String toString()
	{	return "\t\t\t" +
	        " Customer ID :" + key   + "\n" +
	        " Name    : " + name    + "\n" +
	        " Address : " + billingAddress + "\n" +
	        " Ship Address : \n" + shipAddress  + "\n";
	}

	private String					key;
	private String					name;
	private Address					billingAddress;
	private Address					shipAddress;
	private BillingTerms			terms = BillingTerms.get("NET 30");
	private boolean					isActive = true;

	private static class IDGenerator
	{	public static String nextID()
	    {	 return ID_PREFIX + Long.toString(nextID++);
	    }

	    private static long             nextID   = 5001;
	    private final static String     ID_PREFIX = "C";
	}

	private static HashMap allCustomers = new HashMap();

	public static final void close(ObjectOutputStream out) throws IOException
	{	out.writeLong(IDGenerator.nextID);
		out.writeObject(allCustomers);
	}

    public boolean isActive()
    {	return isActive;
    }

	public static final void open(ObjectInputStream in) throws IOException, ClassNotFoundException
	{	IDGenerator.nextID = in.readLong();
		allCustomers = (HashMap)in.readObject();
        //System.out.println(IDGenerator.nextID);
	}

	public static Iterator customers()
	{	return allCustomers.values().iterator();
	}

    public static Iterator keys()
    {	return allCustomers.keySet().iterator();
    }

	public static void dump()
	{	System.out.println("\n\n\n Customers \n\n\n");
		System.out.println(allCustomers);
	}

	public static Customer get(String customerNumber)
	{   Customer customer =  (Customer)allCustomers.get(customerNumber);
	    if(customer == null || customer.isActive) return customer;
        return null;
	}

	public static void remove(String key)
        {   Customer c = (Customer)allCustomers.get(key);
    	    if(c != null)c.isActive = false;
        }


	private static Customer genCustomer(String customer) // not used now. A parser.
	{   StringTokenizer tok = new StringTokenizer(customer, "~");
	    String name = tok.nextToken();
	    Address billingAddress = new Address ( name,    // name
		                        			  tok.nextToken(),    // street
		                        			  tok.nextToken(),    // city
		                        			  tok.nextToken(),    // state
		                        			  tok.nextToken()
		                        			);
	    Address shippingAddress = new Address ( tok.nextToken(),    // name
		                        			  tok.nextToken(),    // street
		                        			  tok.nextToken(),    // city
		                        			  tok.nextToken(),    // state
		                        			  tok.nextToken()	  // zip
		                        			);
		BillingTerms billingTerms = BillingTerms.get("NET 10");
		String terms = tok.nextToken().toUpperCase();
		if(terms.indexOf("NET 10") >= 0) billingTerms = BillingTerms.get("NET 10");
		else if(terms.indexOf("NET 30") >= 0) billingTerms = BillingTerms.get("NET 30");
		else if(terms.indexOf("EOM 10") >= 0) billingTerms = BillingTerms.get("EOM 10");

		return new Customer(name, billingAddress, shippingAddress, billingTerms);
	}

}