package orderprocessing;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.io.Serializable;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.BufferedWriter;
import java.io.Writer;

/** This class represents a single order from a single customer.
* <p>
*	Additionally, this class maintains a collection of all Orders.
*/
public class Order implements Serializable
{	/** Create an Order for a particular customer and a sequence of items ordered
	* @param customer The customer ordering these items
    * @param shipAddress The address the order should be shipped to.
	* @param orderedItems A ArrayList of OrderItems representing the items ordered
	* @param orderDate The date that this order was placed
	*/
	public Order
	(	Customer customer,
		Address shipAddress,
		ArrayList orderedItems,
		CalculationCalendar orderDate
	)
	{	this(customer, shipAddress, null, orderedItems, orderDate, null, null);
		// uses default ship date
	}

	/** Create an Order for a particular customer and a sequence of items ordered
	* @param customer The customer ordering these items
	* @param shipAddress The Shipping Address if different from the customer's default
        * @param billingTerms The billing terms for order if not the customer's default.
	* @param orderedItems A ArrayList of OrderItems representing the items ordered
	* @param orderDate The date that this order was placed
	* @param requestedShippingDate Requested shipping date by customer
	* @param cancelDate Requested cancelation date for this order.
	*/
	public Order
	( Customer customer,
	  Address shipAddress,
          BillingTerms billingTerms,
	  ArrayList orderedItems,
	  CalculationCalendar orderDate,
          CalculationCalendar requestedShippingDate,
          CalculationCalendar cancelDate
	)
	{ key = IDGenerator.nextID();
    	  this.customer    = customer;
	  this.shipAddress  = shipAddress;
	  this.orderedItems   = (ArrayList) orderedItems.clone();
	  this.orderDate   = orderDate;
	  this.terms = billingTerms;

	  if(this.terms == null) this.terms = customer.billingTerms();

	  CalculationCalendar earlyShipDate = this.orderDate.incrementedBy(CalculationCalendar.DATE, MIN_SHIP_DELAY);
	  this.requestedShippingDate = earlyShipDate.max(requestedShippingDate);

	  CalculationCalendar earlyCancelDate = this.requestedShippingDate.incrementedBy(CalculationCalendar.DATE, MIN_CANCEL_DELAY);
	  this.cancelDate = earlyCancelDate.max(cancelDate);

	  allOrders.put(key, this);
	}

    // accessors
	public Customer customer(){return customer;}
	public Address billingAddress(){return customer.billingAddress();}
	public Address shippingAddress(){return shipAddress;}
	public Iterator orderedItems(){return orderedItems.iterator();}

        public int size()
        {	return orderedItems.size();
        }
	public CalculationCalendar orderDate(){return orderDate;}
	public CalculationCalendar requestedShippingDate(){return requestedShippingDate;}
	public CalculationCalendar cancelDate(){return cancelDate;}
	public boolean valid(){return validOrder;}
	public String key() {return key;}
	public BillingTerms billingTerms(){return terms;}
	public boolean isCancelled(){ return cancelled;}


	/** return the extended total dollar value of this order
	* @return the total value
	*/
	public double total()
	{	Iterator items = orderedItems();
		double total = 0.0;
		while(items.hasNext())
		{	OrderItem ordered = (OrderItem)items.next();
			total += ordered.quantity() * ordered.unitPrice();
		}
		return total;
	}

    // mutators
	public void cancelOrder(){ cancelled = true;}

	public void billingTerms(BillingTerms billingTerms)
	{	 terms = billingTerms;
	}

    public static void formatAll(BufferedWriter file) // NOTE, NOT in order number order.
    {	Iterator orders = orders();
    	Order order = null;
        while (orders.hasNext())
        {	try
        	{	order = (Order)orders.next();
            	order.format(file);
    			file.newLine();
            	file.newLine();
        	}
            catch(IOException ex)
            {	System.out.println("Failed to write order "+order.key() +"  " + ex);
            }
        }
    }

    public void format(BufferedWriter file) throws IOException
    {	// to do
    }

	public void shippingAddress(Address address){shipAddress = address;}

    // other methods
	public void display()
	{	System.out.println("\t\t\t------- Order ------- " + key);
	    customer.display();

	    System.out.println("\n\n\tQty\tItem No. " +
	            "\tDescription\t\tUnit Cost\n");

		Iterator items = orderedItems();
		while(items.hasNext())
		{	((OrderItem)items.next()).display();
		}

		terms.display();

		System.out.println("Order    Date: " + orderDate.getTime());
		System.out.println("Customer Shipping information:");
		System.out.println(shippingAddress());
		System.out.println(" Ship Date: " +
		                   requestedShippingDate.getTime());
		System.out.println(" Cancel Date: " +
		                   cancelDate.getTime());

	    System.out.println("---------------------");
	}

	public boolean equals(Object other)
	{	if(!(other instanceof Order))return false;
		return key.equals(((Order)other).key);
	}

	public int hashcode()
	{	return key.hashCode()*997;
	}

	public String toString() {return "Order with id: " + key;}

	private String                      key;
 	private Customer                    customer;
	private Address                     shipAddress;
	private ArrayList                      orderedItems;       // ArrayList of OrderItems
	private CalculationCalendar         orderDate;
	private CalculationCalendar         requestedShippingDate;
	private CalculationCalendar         cancelDate;// if null order is still open
	private BillingTerms                terms = BillingTerms.get("NET 10");

	private boolean						cancelled = false;
	private boolean						validOrder = true;

	private static final int MIN_SHIP_DELAY = 7;
	private static final int MIN_CANCEL_DELAY = 10;

	private static class IDGenerator
	{	public static String nextID()
	    {	return ID_PREFIX + Long.toString(nextID++);
	    }

	    private static long             nextID = 2001;
	    private final static String     ID_PREFIX = "O";
	}

    // Following is the collection of all Orders.
	private static HashMap allOrders = new HashMap();

    public static Order get(String key)
    {	return (Order)allOrders.get(key);
    }

	public static final void close(ObjectOutputStream out) throws IOException
	{	out.writeLong(IDGenerator.nextID);
		out.writeObject(allOrders);
	}

	public static final void open(ObjectInputStream in) throws IOException, ClassNotFoundException
	{	IDGenerator.nextID = in.readLong();
		allOrders = (HashMap)in.readObject();
	}

	public static Iterator orders()
	{	return allOrders.values().iterator();
	}

	public static Iterator keys()
	{	return allOrders.keySet().iterator();
	}

	public static void dump()
	{	System.out.println("\n\n\n Orders \n\n\n");
		System.out.println(allOrders);
	}
}

/* notes shipdate  after order date + 7
	cancel date after ship date + 10
*/